#
# ident	"@(#)domain.tcl	1.65	99/08/23 SMI"
#
# Copyright (c) 1999 by Sun Microsystems, Inc.
# All rights reserved.
#
#	(1) This file contains all Tcl entry points for domain tool commands:
#
#		domain_create
#		domain_remove
#		domain_rename
#		domain_status
#		domain_history
#
#	domain.tcl is loaded by scotty, and each function is executed
#	as a Unix command, the name and arguments being passed to a
#	scotty shell through the magic of proxy programs and symbolic
#	links.
#
#
# domain_config positioning globals.
# NOTE: THESE ARE DUPLICATED in the following files:
#       pd/scotty*/library/ssp_init.tcl
#       cmd/domain_tools/domain.tcl
#       cmd/startup/ssp_startup.tcl
#
# name : platform : machine : osversion : idninfo : sysbds
#
set gdcfg_name_position		0
set gdcfg_platform_position	1
set gdcfg_physical_position	2
set gdcfg_osver_position	3
set gdcfg_idninfo_position	4
set gdcfg_sysbds_position	5

# Initialize global variables.
#
set domain_tools_debug 0
set EEPROM_DIR eeprom_save
set SSP_USER ssp
set D_PLATFORM_TYPE unknown
set domain_config_ok 1

rename error error.real

proc error {s} \
{
  upvar handle handle
  upvar progname progname
  upvar domain_name domain_name

  if { [info exists handle] } {
    if { [string length $handle] } {
      $handle destroy
    }
    #
    # If we are here, and the caller is domain_create, then we know
    # that the ssp_to_domain_hosts entry was added (handle variable
    # exists, which is set right after S.T.D.H call) - remove it.
    #
    if { [info exists progname] && [info exists domain_name] } {
      if { [string match $progname "domain_create" ] } {
        rm_ssp_to_domain_hosts_entry $domain_name
      }
    }
  }

  error.real "$s"
}

##############################################################################
#
# set_default_platform_type:  Set default platform type.
#
# inputs:
#	none
#
# outputs: 
#	D_PLATFORM_TYPE is set if cb_config can be read, otherwise error
#	is raised, aborting command.
#
##############################################################################
proc set_default_platform_type {} \
{
  global D_PLATFORM_TYPE

  upvar progname progname

  if [ catch { get_cb_config } cb_config ] {
    error "$progname: Error reading cb_config file."
  } elseif { ![string length $cb_config] } {
    error "$progname: cb_config file is empty."
  }

  set D_PLATFORM_TYPE [lindex [lindex $cb_config 0] 1]
}

##############################################################################
#
# check_ssp_env:  Verify SSP env variables are defined.
#
# inputs:
#	none
#
# outputs: 
#	error raised if any of the required variables are not defined.
#
##############################################################################
proc check_ssp_env {} \
{
  upvar progname progname
  global env

  foreach var {SSPOPT SSPVAR SSPETC SSPLOGGER} {
    if {[info exists env($var)] == 0} {
      error "$progname: Environment variable '$var' must be set."
    }
  }
}

##############################################################################
#
# check_domain_history:  Verify domain_history file exists, create it if it
#			 doesn't.
#
# inputs:
#	none
#
# outputs: 
#	error raised if file cannot be created.
#
##############################################################################
proc check_domain_history {} \
{
  upvar progname progname
  global env

  set dhf $env(SSPVAR)/.ssp_private/domain_history

  if { ![file exists $dhf] } {
    if [ catch {open $dhf {RDWR CREAT}} fileId ] {
      error "$progname: $fileId"
    } else {
      # success
      close $fileId
    }
  }
}

##############################################################################
#
# open_snmp:  Open commnication session with SNMP agent.
#
# inputs:
#	none
#
# outputs: 
#	handle	Handle to new session.
#
##############################################################################
proc open_snmp {} \
{
  global env

  upvar progname progname

  if [ catch { snmp session } handle ] {
    puts "$progname: SNMP session error: $handle"
    return ""
  }
  return $handle
}

##############################################################################
#
# eeprom_index:  Get used/unused eeprom file index.
#
# inputs:
#	file_exists	establishes criteria for index search.  If true,
#			eeprom_index finds the first eeprom.image.<INDEX>
#			that exists; if false, eeprom_index returns the 
#			index of the first eeprom.image.<INDEX> file that 
#			does not exist.  Both searches start at zero.
#
#	idx		pass-by-reference variable that will contain the
#			index value if the search is successful.
#
# outputs: 
#	1 on success, 0 on failure.
#
##############################################################################
proc eeprom_index { file_exists idx } \
{
  global env
  global EEPROM_DIR

  upvar $idx retval

  set handle [open_snmp]
  if { [string length $handle] == 0 } {
    return 0
  }

  if [catch {lindex [lindex [$handle get confNumSysBrd.0] 0] 2} num_sysbrds ] {
     error "Error getting system board count from SNMP agent"
  }

  $handle destroy

  set e $env(SSPVAR)/.ssp_private/$EEPROM_DIR
  for {set i 0} {$i < $num_sysbrds} {incr i} {
    if { $file_exists } {
      if { [file exists $e/eeprom.image_$i] } {
	    break
      }
    } else {
      if { ![file exists $e/eeprom.image_$i] } {
	    break
      }
    }
  }
  if { $i < $num_sysbrds } {
    set retval $i 
    return 1
  }
  return 0
}

##############################################################################
#
# free_eeprom_file:  Deallocate one or more eeprom.image files assigned to
#		     domain history entries.
#
# inputs:
#	domain_name	If non-zero length, free eeprom image file for
#			specified domain, otherwise, free all eeprom
#			image files.
# outputs: 
#	none
#
##############################################################################
proc free_eeprom_file { domain_name } \
{
  global env
  global EEPROM_DIR
  upvar progname progname
  global gdcfg_name_position

  set e $env(SSPVAR)/.ssp_private/$EEPROM_DIR

  if { [string length $domain_name] == 0 } {
    set h_llist [get_domain_history]
  } else {
    set h_llist [get_domain_history_entry $domain_name]
  }

  for {set i 0} {$i < [llength $h_llist]} {incr i} {
    set d_name [lindex [lindex $h_llist $i] $gdcfg_name_position]
    if { [file exists $e/eeprom.image.$d_name] } {
      if { [eeprom_index 0 idx] } {
        exec mv $e/eeprom.image.$d_name $e/eeprom.image_$idx
      } else {
	    set pid_name $e/eeprom.image.pid[pid]
	    puts "$progname: Warning: Could not find unused eeprom slot."
	    puts "$progname: Renaming eeprom.image.$d_name to $pid_name"
	    exec mv $e/eeprom.image.$d_name $pid_name
      }
    }
  }
}

proc check_user {} \
{
  global SSP_USER
  upvar progname progname

  #
  # Get SSP name from OS environment
  #
  set input [open "|/usr/xpg4/bin/id -u -n" r]
  set user_name [gets $input]
  close $input
  if { ![string match $user_name $SSP_USER ] } {
    error "Please execute $progname as user $SSP_USER"
  }
}

##############################################################################
#
# domain_create_usage:  Print usage info for domain_create command and
#			raise error.
#
# inputs:
#	none
#
# outputs: 
#	none
#
##############################################################################
proc dc_usage {} \
{
  upvar progname progname
  error "Usage: $progname -d domain_name \[-b board_numbers\] \[-o OS_version\]\n\t\t\t\
			\[-p platform_name\] \[-t platform_type\]"
}

##############################################################################
#
# domain_create:  Create a domain - setup files and directories, use xf_post
#		  to configure the system boards, and add an entry to the
#		  domain_config file.
#
# inputs:
#	args	list of comand line arguments, passed as an array.
#
# outputs: 
#	0 on success, 1 on failure.
#
##############################################################################
proc domain_create { args } \
{
  global env
  global EEPROM_DIR
  global D_PLATFORM_TYPE
  global domain_tools_debug
  global domain_config_ok
  global gdcfg_sysbds_position
  global gdcfg_osver_position
  global gdcfg_physical_position

  set progname "domain_create"

  check_user
  check_ssp_env
  set_default_platform_type
  check_domain_history

  #
  # "Option found" flags - initialize.
  #
  set d_op 0
  set b_op 0
  set o_op 0
  set p_op 0
  set t_op 0

  #
  # Default argument values.
  #
  set domain_name ""
  set board_numbers ""
  set os_version ""
  set platform_name ""
  set platform_type $D_PLATFORM_TYPE
  set quiet 0

  #
  # Find number of arguments, process/validate each.
  #
  set argcnt [llength $args]
  set ac 0
  set board_mode 0
  while {$ac < $argcnt} {
    #
    # Process command arguments.  Each option must have an argument; if not,
    # print usage information and raise error.
    #
    switch -glob -- [lindex $args $ac] {
      -d { if {[incr ac] >= $argcnt} { dc_usage } ;
		set domain_name [lindex $args $ac] ;
		incr d_op ; set board_mode 0 }
      -b { if {[incr ac] >= $argcnt} { dc_usage } ;
		set board_numbers [lindex $args $ac] ;
		incr b_op ; incr board_mode }
      -o { if {[incr ac] >= $argcnt} { dc_usage } ;
		set os_version [lindex $args $ac] ;
		incr o_op ; set board_mode 0 }
      -p { if {[incr ac] >= $argcnt} { dc_usage } ;
		set platform_name [lindex $args $ac] ;
		incr p_op ; set board_mode 0 }
      -t { if {[incr ac] >= $argcnt} { dc_usage } ;
		set platform_type [lindex $args $ac] ;
		incr t_op ; set board_mode 0 }
      -q { set quiet 1 }
      -* { dc_usage }
      default { if { $board_mode } {
		  set board_numbers [concat $board_numbers [lindex $args $ac]]
		} else { dc_usage } }
    }
    incr ac
  }
  #
  # If user specified board list as "a,b,c...", replace commas with spaces.
  #
  regsub -all , $board_numbers " " board_numbers

  #
  # Check that non-zero-length domain_name was specified.
  #
  if { ![string length $domain_name] } {
    puts "Must specify domain name."
    dc_usage
  }

  #
  # Check for illegal characters in domain name.
  #
  if { [string first ":" $domain_name] >= 0 } {
    error  "$progname: Character ':' not allowed in domain name."
  }

  #
  # Check to see whether domain_config file exists.  If it doesn't,
  # skip all checks of that file, as well as any checks with the 
  # snmp daemon since without a domain_config file, that daemon 
  # would not be running.
  #
  if { [file exists $env(SSPVAR)/.ssp_private/domain_config] && \
	[file size $env(SSPVAR)/.ssp_private/domain_config] > 0 } {
    set domain_config_ok 1
  } else {
    set domain_config_ok 0
  }

  #
  # Check that domain_name is not already in use.
  #
  if { $domain_config_ok } {
    if { [llength [get_domain_entry $domain_name]] } {
      error "$progname: Domain '$domain_name' already exists."
    }
  }

  #
  # Get history entry, if any.
  #
  set h_list [lindex [get_domain_history_entry $domain_name] 0]

  #
  # Check options specified.  If any are missing, try to get info from 
  # history entry for the specified domain.
  #
  if {!$b_op} {
    if {[llength $h_list]} {
      set board_numbers [lindex $h_list $gdcfg_sysbds_position]
    } else {
      error "$progname: No history info for domain\
	    '$domain_name'. Board numbers required."
    }
  }
  if {!$o_op} {
    if {[llength $h_list]} {
      set os_version [lindex $h_list $gdcfg_osver_position]
    } else {
      error "$progname: No history info for domain\
	    '$domain_name'. OS version required."
    }
  }
  if {!$p_op} {
    if {[llength $h_list]} {
      set platform_name [lindex $h_list $gdcfg_physical_position]
    } else {
        error "$progname: No history info for domain\
	      '$domain_name'. Platform name required."
    }
  }
  #
  # Initialize IDN fields.
  #
  set idn_info -1.0
 
  #
  # Read cb_config file in order to verify platform name and type.
  #
  if [ catch { get_cb_config } cb_config ] {
    error "$progname: Error reading cb_config file."
  } elseif { ![string length $cb_config] } {
    error "$progname: cb_config file is empty."
  }

  for {set i 0} {$i < [llength $cb_config]} {incr i} {
    if { [string match $platform_name [lindex [lindex $cb_config $i] 0]] } {
      break
    }
  }

  if { $i >= [llength $cb_config] } {
    error "$progname: Platform '$platform_name' not found."
  }

  #
  # Check that platform_type is one we support.
  #
  if { ![string match $platform_type [lindex [lindex $cb_config $i] 1]]} {
      error "$progname: Platform type '$platform_type' not supported."
  }

  #
  # Check OS version support.
  #
  set os_dirs [split $os_version .]
  set num_os_dirs [llength $os_dirs]

  if { $num_os_dirs < 2 || [lsearch -exact $os_dirs ""] >= 0 } {
    error "$progname: Invalid OS version: '$os_version'"
  }

  # Release number the user has input
  set d_major [ lindex $os_dirs 0 ]
  set d_minor [ lindex $os_dirs 1 ]
  if { $num_os_dirs > 2 } {
    set d_patch [ lindex $os_dirs 2 ]
  } else {
    set d_patch 0
  }
  set d_rel $d_major$d_minor$d_patch

  # Minimum release we support: 2.5.0
  set major 2
  set minor 5
  set patch 0
  set rel $major$minor$patch

  # Compare them, anything less than $rel is invalid!
  if { $d_rel < $rel } {
    error "$progname: OS version '$os_version' not supported."
  }

  # Be sure major dir exists
  set rel_dir $env(SSPOPT)/release/$platform_type/$major
  if { ![file isdirectory $rel_dir] } {
    error "$progname: $rel_dir directory does not exist."
  }

  # Be sure user's $d_rel exists
  set rel_dir $env(SSPOPT)/release/$platform_type/$d_major/$d_minor
  if { ![file isdirectory $rel_dir] } {
    puts "$progname: warning - $rel_dir directory does not exist."
  }

  #puts "DEBUG: rel_dir is $rel_dir"

  #
  # Set HOSTNAME to the platform specified, or recovered from
  # the domain_history file.  This enables SNMP to work correctly.
  #
  set env(SUNW_HOSTNAME) $platform_name

  #
  # Get SSP name from OS environment
  #
  set input [open "|uname -n" r]
  set ssp_name [gets $input]
  close $input

  set handle [open_snmp]
  if { [string length $handle] == 0 } {
     error "Error connecting to SNMP agent"
  }

  #
  # Get the confSysBrdList SNMP MIB value, and from that list, make a list of
  # of boards present in the system.
  #
  if [catch {lindex [lindex \
			[$handle get confSysBrdList.0] 0] 2} sys_board_info ] {
     error "Error getting system board list from SNMP agent"
  }

  set sysb_list [split $sys_board_info -]

  #
  # Check that board numbers specified exist in the system, and are not already
  # in use by another domain with the same platform name.
  #

  set board_numbers [string trim $board_numbers]
  if { [string length $board_numbers] } {

    #
    # Split proposed boards argument into a list.
    #

    set newb_list [split $board_numbers]

    if { $domain_config_ok } {
      #
      # Retrieve entire contents of domain_config file.
      #
      set d_config [get_domain_config]
    }

    #
    # Loop through new board list and make sure a. it is physically present
    # in the system, and b. it is not in use by another active domain.
    #

    for {set i 0} {$i < [llength $newb_list]} {incr i} {

      set brd [lindex $newb_list $i]

      #
      # Range check
      #
      if { $brd < 0 || $brd > 15 } {
	error "$progname: Board numbers must be between\
               0 and 15, inclusive."
      }

      #
      # Check that requested board is physically present.
      #
      if { ! $quiet } {
	  if { [lsearch -exact $sysb_list $brd] < 0 } {
	      puts stdout "\n$progname: Board $brd does not exist."
	      puts -nonewline stdout "\nProceed anyway? (y) "
	      set answer [gets stdin]
	      if { [string match {[Nn]*} $answer] } {
		  error "$progname: Abort."
	      }
	  }
      }

      if { $domain_config_ok } {
        #
        # Check that board is not in use by another domain.
        #

        for {set j 0} {$j < [llength $d_config]} {incr j} {
          set d_list [lindex $d_config $j]

          #
          # Check for a platform name match.
          #

          if { [string match $platform_name [lindex $d_list $gdcfg_physical_position]] } {

	    #
	    # Get board list for this domain.
	    #

            set domb_list [split [string trim [lindex $d_list $gdcfg_sysbds_position]]]
            if { [lsearch -exact $domb_list $brd] >= 0 } {
	      error [format "%s: Board %d already used by domain '%s'" \
			$progname $brd [lindex $d_list 0]]
	    }
	  }
        }
      }
    }
  }

  set p_etc $env(SSPVAR)/etc/$platform_name
  set d_etc $p_etc/$domain_name

  #
  # BEGIN:  Critical section
  #
  trap SIG_IGN SIGINT
  trap SIG_IGN SIGQUIT

  #
  # Following line should not be moved without a good reason.  From
  # this point forward, we must remove the ssp_to_domain_hosts entry
  # for this domain if the command is aborted (error call) for some
  # reason.
  #
  catch {put_ssp_to_domain_hosts_entry [list [list $domain_name $ssp_name]]} status

  #
  # Create required files and directories.
  #
  if { [catch {dc_file_mgr "create" $platform_name $domain_name "" $platform_type} status] } {
        trap SIGINT
        trap SIGQUIT
        error $status
  }

  #
  # Install eeprom image.
  #
  set e $env(SSPVAR)/.ssp_private/$EEPROM_DIR

  #
  # Make sure eeprom directory exists.
  #
  if { ![file isdirectory $e] } {
    exec rm -f $e
  }

  if { ![file exists $e] } {
    exec mkdir $e
    exec chmod 755 $e
  }

  if {[llength $h_list]} {
    if [ catch { exec cp $e/eeprom.image.$domain_name \
        $d_etc/eeprom.image } ret_cp ] {
    	catch {dc_file_mgr "remove" $platform_name $domain_name} status
        trap SIGINT
        trap SIGQUIT
        error "Missing $e/eeprom.image.$domain_name!"
    }
  } else {

    #
    # Reclaim eeprom file from history file if same domain name exists.
    # Else find next unused eeprom file.
    # install into etc/domain area. 

   if [ file exists $e/eeprom.image.$domain_name ] {
       if [catch {exec cp $e/eeprom.image.$domain_name \
          $d_etc/eeprom.image } ret_cp ] {
          catch {dc_file_mgr "remove" $platform_name $domain_name} status
          trap SIGINT
          trap SIGQUIT
          error "Error, check permission of file\
	    $e/eeprom.image.$domain_name !"
       }
    } elseif { [eeprom_index 1 i] } {
      #
      # Found an unused file.  Copy it, then rename it with the
      # domain name embedded.
      #
      if [catch {exec cp $e/eeprom.image_$i \
          $d_etc/eeprom.image } ret_cp ] {
    	  catch {dc_file_mgr "remove" $platform_name $domain_name} status
          trap SIGINT
          trap SIGQUIT
          error "Error, check permission of file\
	    $e/eeprom.image_$i!"
      }
      if [catch {exec mv $e/eeprom.image_$i \
          $e/eeprom.image.$domain_name } ret_cp ] {
          puts stdout "Warning can't save $e/eeprom.image_$i. \n \
		       Check permission of file !"
      }

    } else {
        #
        # No eeprom image found for this domain, exit. 
        #
        catch {dc_file_mgr "remove" $platform_name $domain_name} status
        trap SIGINT
        trap SIGQUIT
        error "$progname: Error! No eeprom file for this domain: $domain_name"
    }
  }

  #
  # Add the entry to domain_config.  Create a list of a list of 1 to pass
  # to put_domain_entry().
  #
  set d_new [list [list $domain_name $platform_type $platform_name $os_version \
					$idn_info "$board_numbers"]]
  if { $domain_config_ok } {
    if [ catch { put_domain_entry $d_new } ret_ent ] {
      error "put_domain_entry : $ret_ent"
    }
  } else {
    #
    # We should only get into this section if we need to create the
    # domain_config file the first time. We would normally need to
    # set OVERRIDEFAD but there is code in the fad interface that
    # defaults to the local SSP host, so no need to for now.
    #
    set d_new [ format "{{{%s} {%s} {%s} {%s} {%s} {%s}}}" \
		 $domain_name $platform_type $platform_name \
		 $os_version  $idn_info "$board_numbers"]
    if [ catch { put_domain_config $d_new } ret_dc ] {
      error "put_domain_config : $ret_dc"
    }
  }

  #
  # Verify creation.
  #
  if { ![llength [get_domain_entry $domain_name]] } {
    catch {dc_file_mgr "remove" $platform_name $domain_name} status
    error "$progname: Error creating domain '$domain_name'."
  }

  #
  # Remove entry from history file, if it was found.
  #
  if {[llength $h_list]} {
    catch {rm_domain_history_entry $domain_name} status
  }

  if {$domain_tools_debug > 0} {
    puts stdout [format "domain_name   : '%s'" $domain_name]
    puts stdout [format "board_numbers : '%s'" $board_numbers]
    puts stdout [format "os_version    : '%s'" $os_version]
    puts stdout [format "platform_name : '%s'" $platform_name]
    puts stdout [format "platform_type : '%s'" $platform_type]
    puts stdout [format "idn_info      : '%s'" $idn_info]
  }

  $handle destroy

  #
  # Clear all of our boards from hw masks, this prevents certain
  # types of HW SMD mask confusion from being carried over into the new
  # domain.
  #
  if [ catch { rm_brds_from_masks $board_numbers } err ] {
      set fmt [gettext "ERROR: unable to clear boards %s from HW masks"]
      puts stderr [format "$progname: $fmt" "$board_numbers" ]
  }

  #
  # Reset the signature block of the first proc on the new domain.
  #
  set procnum [expr [lindex $board_numbers 0] * 4]
  set old_domain $env(SUNW_HOSTNAME)
  set env(SUNW_HOSTNAME) $domain_name
  if { [catch {exec sigbcmd -S -P $procnum }] != 0} {
      	puts stderr "$progname: ERROR: Unable to reset processor $procnum"
  }
  set env(SUNW_HOSTNAME) $old_domain

  puts stdout "Domain : $domain_name is created !"

  #
  # END:  Critical section
  #
  trap SIGINT
  trap SIGQUIT
}

##############################################################################
#
# domain_remove_usage:  Print usage info for domain_remove command and
#			raise error.
#
# inputs:
#	none
#
# outputs: 
#	none
#
##############################################################################
proc domain_remove_usage {} {
  upvar progname progname
  error "Usage: $progname -d domain_name"
}

##############################################################################
#
# domain_remove:  Remove a domain - remove associated files and directories,
#		  deallocate system boards, remove entry from the
#		  domain_config file, clear the hardware masks for each
#		  board in the domain, and add entry to domain_history file.
#
# inputs:
#	args	list of comand line arguments, passed as an array.
#
# outputs: 
#	0 on success, 1 on failure.
#
##############################################################################
proc domain_remove { args } \
{
  global env
  global EEPROM_DIR
  global domain_tools_debug
  global gdcfg_physical_position
  global gdcfg_sysbds_position

  set progname "domain_remove"

  check_user
  check_ssp_env
  set_default_platform_type
  check_domain_history

  #
  # Default argument values.
  #
  set domain_name ""
  set quiet 0
  set removeDirs -1

  #
  # Find number of arguments, process/validate each.
  #
  set argcnt [llength $args]
  set ac 0
  while {$ac < $argcnt} {
    #
    # Process command arguments.  Each option must have an argument; if not,
    # print usage information and raise error.
    #
    switch -exact -- [lindex $args $ac] {
      -d { if {[incr ac] >= $argcnt} { domain_remove_usage } ;
		set domain_name [lindex $args $ac]
	 }

      -q { set quiet 1 }

	  -y { set removeDirs 1 }

	  -n { set removeDirs 0 }

      default { domain_remove_usage }
    }
    incr ac
  }

  #
  # Check that non-NULL domain_name was specified.
  #
  if { ![string length $domain_name ] } {
    puts "Must specify domain name."
    domain_remove_usage
  }

  #
  # Check that domain_name exists; if so, create list from data.
  #
  if [ catch { get_domain_entry $domain_name } d_list ] {
	# No domain_config file.
	error "$progname: No such domain: '$domain_name'"
  }

  if { ![llength $d_list] } {
    error "$progname: No such domain: '$domain_name'"
  }
  set env(SUNW_HOSTNAME) $domain_name
  set platform_name [lindex [lindex $d_list 0] $gdcfg_physical_position]
  set d_adm $env(SSPVAR)/adm/$domain_name
  set d_etc $env(SSPVAR)/etc/$platform_name/$domain_name


  set handle [open_snmp]
  if { [string length $handle] == 0 } {
    puts stderr "$progname: WARNING - cannot obtain configured board list for domain from snmpd."
    puts stderr "$progname: Using domain config file board list..." 
    set brdlist  [lindex [lindex $d_list 0] $gdcfg_sysbds_position]
  } else {
     $handle walk x "domainIndex domainName domainIDNId" {
        set indx [lindex [lindex $x 0] 2]
        set host [lindex [lindex $x 1] 2]
        if { $host == $domain_name } {
	     set idnid [lindex [lindex $x 2] 2]
	     if { $idnid != -1 } {
			puts stderr \
"$progname: Domain ($domain_name) is linked to an IDN Network.\n\
\t\tDomain must be unlinked before it can be removed."
			error "\t\tSee domain_unlink(1M)."
	     }
             # Got our domain name
             set brdlist [lindex [lindex [$handle get domainSysBrdConfig.$indx] 0] 2]
             set brdlist [ split $brdlist - ]
             break
        }
     }
  }

  # At this point it is possible that brdlist is empty.  This can happen
  # if the domain has been created, but has NOT been booted.  If this
  # does happen, then it is okay to ignore the clr_domain_brd() call
  # later in this routine.

  #
  # Query the user if he/she wants that all domain data be removed.
  #
  if { ! $quiet } {
      puts stdout "\n$progname: The following subdirectories contain\
          domain specific\n information such as messages files,\
          configuration files, and hpost dump files.\n You may choose\
          to keep these directories if you still need this information.\n\
          This domain may be recreated with or without this information being saved."
      puts stdout "\t\t$d_adm"
      puts stdout "\t\t$d_etc"
      puts -nonewline stdout "\nKeep directories (y/n)? "
      while { $removeDirs == -1 } {
         set answer [gets stdin]
         if { $answer == "Y" || $answer == "y" } {
            set removeDirs 0
            break
         } elseif { $answer == "N" || $answer == "n" } {
            set removeDirs 1
            break
         } else {
            puts -nonewline stdout "$progname: Keep directories ? (y/n) "
         }
      }
  }

	#
	# Make sure domain is not up and running.  Save current
	# domain/HOSTNAME, put proposed domain name in HOSTNAME,
	# check_host, then restore HOSTNAME to original value.
	#
  set old_domain $env(SUNW_HOSTNAME)
  if { ! $quiet } {
	set env(SUNW_HOSTNAME) $domain_name

	if [ catch { check_host -q } result ] {
		# restore HOSTNAME
		set env(SUNW_HOSTNAME) $old_domain
		if { [string length $result] } {
			error "$progname: $result"
		} else {
			error "$progname: Error checking domain '$domain_name'"
		}
	}

	set env(SUNW_HOSTNAME) $old_domain
	if { $result == 0 } {
		# Host is UP
		puts "\n$progname: Domain '$domain_name' is UP, must be shutdown before removing."
		puts -nonewline stdout "Continue  anyway (y/n)? "
		while { $result == 0 } {
			set answer [gets stdin]
			if { $answer == "Y" || $answer == "y" } {
				break
			} elseif { $answer == "N" || $answer == "n" } {
				error "$progname: Abort."
			} else {
				puts -nonewline stdout "$progname: Continue anyway (y/n)? "
			}
		}
	}
  }

  #
  # BEGIN:  Critical section
  #
  trap SIG_IGN SIGINT
  trap SIG_IGN SIGQUIT



  #
  # check for arbstop dumps that might be running
  # by waiting for the lock.  Execution pauses here until lock
  # is obtained.  There is no release.
  #
  # get the postdump lock file
  #
  set lock [ lock_ssp_file "eddpostdump.lock 0" ]


  #
  # Remove old history entry for this domain, if any, then add new entry.
  #
  if [ catch { rm_domain_history_entry $domain_name } rm_ret ] { 
    error "rm_domain_history_entry : $rm_ret"
  }

  if { [ catch { put_domain_history_entry $d_list } pt_ret ] } { 
     error "put_domain_history_entry : $pt_ret"
  }
 
  #
  # clear each board's hardware masks
  #
  # NOTE - we must do this before we remove the entry from the
  # domain_config file since rm_brds_in_mask_from_smds talks to cbs which
  # requires the entry for the domain to be in the domain_config file.
  # Save and restore SUNW_HOSTNAME as above.
  #
  set old_domain $env(SUNW_HOSTNAME)
  set env(SUNW_HOSTNAME) $domain_name

  # Pick up the board list length and if it is zero skip the
  # rm_brds_from_masks() call.  No need to reset CPUs, this is done
  # by rm_boards_in_mask_from_smds() in libssphw.
  set brd_list_len [llength $brdlist]

  if { $brd_list_len > 0 } {
    # call a single clear for all boards in domain.
    if [ catch { rm_brds_from_masks $brdlist } err ] {
      puts stderr "Failed to clear hardware domain masks"
      if { ! $quiet } {
             puts -nonewline stdout "$progname: Proceed with domain_remove? (y/n)"
	     set answer [gets stdin]
	             if { ![string match {[Yy]*} $answer] } {
                     error "$progname: Abort."
                     }
      }
    }
  }

  set env(SUNW_HOSTNAME) $old_domain

  #
  # Remove entry from domain_config.
  #
  if { [ catch { rm_domain_entry $domain_name } rm_ent ] } {
     error "rm_domain_entry : $rm_ent"
  }
 
  #
  # Save eeprom.image data to private area.
  #
  set d_etc $env(SSPVAR)/etc/$platform_name/$domain_name

  #
  # Make sure eeprom directory exists.
  #
  set e $env(SSPVAR)/.ssp_private/$EEPROM_DIR
  if { ![file isdirectory $e] } {
    exec rm -f $e
  }
  if { ![file exists $e] } {
    exec mkdir $e
    exec chmod 755 $e
  }

  if [catch {exec cp $d_etc/eeprom.image \
             $e/eeprom.image.$domain_name } ret_cp ] {
     error "$ret_cp.\nDomain_remove is not complete !"
  }

  #
  # Remove files and directories associated with domain.
  #
  if { $removeDirs } {
      catch {dc_file_mgr "remove" $platform_name $domain_name} status
  }

  catch {rm_ssp_to_domain_hosts_entry $domain_name} status

  # remove domain information from the ssp_resource file.
  catch { ssp_rm_resource $domain_name.domainInterruptVector }
  catch { ssp_rm_resource $domain_name.domainBootProc }
  catch { ssp_rm_resource $domain_name.domainSysBrdConfig }
  catch { ssp_rm_resource $domain_name.domainProcConfig }
  catch { ssp_rm_resource $domain_name.domainABusConfig }
  catch { ssp_rm_resource $domain_name.domainDBusConfig }


  set e $env(SSPVAR)/.ssp_private/$EEPROM_DIR


  #
  # Track down netcon_server PID and kill it.
  #
  set pid_file $env(SSPVAR)/pids/netcon_server-$domain_name.pid
  if { [file exists $pid_file] } {
   if {[ tokenizeFile $pid_file tok " " ] == 1 } {
      set pid [lindex $tok(1) 0]
      if [catch { exec kill -s KILL $pid }] {
        # puts stderr "$progname: Warning! Could not kill netcon_server."
      }
   }
  }

  #
  # Track down obp_helper PID and kill it.
  #
  set pid_file $env(SSPVAR)/pids/obp_helper-$domain_name.pid
  if { [file exists $pid_file] } {
   if {[ tokenizeFile $pid_file tok " " ] == 1 } {
      set pid [lindex $tok(1) 0]
      if [catch { exec kill -s KILL $pid }] {
        # puts stderr "$progname: Warning! Could not kill obp_helper."
      }
   }
  }

  #
  # END:  Critical section
  #
  trap SIGINT
  trap SIGQUIT
  puts stdout "Domain : $domain_name is removed !"

}

##############################################################################
#
# domain_rename_usage:  Print usage info for domain_rename command and
#			raise error.
#
# inputs:
#	none
#
# outputs: 
#	none
#
##############################################################################
proc domain_rename_usage {} {
  upvar progname progname
  error "Usage: $progname -d domain_name -n new_domain_name"
}

##############################################################################
#
# domain_rename:  Rename a domain - change all references to old domain
#		  name to new domain name.
#
# inputs:
#	args	list of comand line arguments, passed as an array.
#
# outputs: 
#	0 on success, 1 on failure.
#
##############################################################################
proc domain_rename { args } \
{
  global env
  global EEPROM_DIR
  global domain_tools_debug
  global gdcfg_physical_position
  global gdcfg_name_position

  set progname "domain_rename"

  check_user
  check_ssp_env
  set_default_platform_type
  check_domain_history

  #
  # Default argument values.
  #
  set domain_name ""
  set new_domain_name ""
  set quiet 0


  #
  # Find number of arguments, process/validate each.
  #
  set argcnt [llength $args]
  set ac 0
  while {$ac < $argcnt} {
    #
    # Process command arguments.  Each option must have an argument; if not,
    # print usage information and raise error.
    #
    switch -exact -- [lindex $args $ac] {
      -d { if {[incr ac] >= $argcnt} { domain_rename_usage } ;
		set domain_name [lindex $args $ac] }
      -n { if {[incr ac] >= $argcnt} { domain_rename_usage } ;
		set new_domain_name [lindex $args $ac] }
      -q { set quiet 1 }
      default { domain_rename_usage }
    }
    incr ac
  }

  #
  # Check that non-NULL domain_names were specified.
  #
  if { ![string length $domain_name ] } {
    puts "Must specify domain name."
    domain_rename_usage
  }
  if { ![string length $new_domain_name ] } {
    puts "Must specify new domain name."
    domain_rename_usage
  }

  #
  # Check for illegal characters in new domain name.
  #
  if { [string first ":" $new_domain_name] >= 0 } {
    error  "$progname: Character ':' not allowed in domain name."
  }

  #
  # Check that domain entry exists.
  #
  set d_list [lindex [get_domain_entry $domain_name] 0]
  if { ![llength $d_list] } {
    error "$progname: No such domain: '$domain_name'."
  }

  #
  # Check that domain_name is not already in use.
  #
  if { [llength [get_domain_entry $new_domain_name]] } {
    error "$progname: Domain '$new_domain_name' already exists."
  }

  set h_list [lindex [get_domain_history_entry $new_domain_name] 0]
  if {[llength $h_list]} {
      if { ! $quiet } {
	  puts stdout \
		  "\n$progname: Warning: History entry for '$new_domain_name' exists."
	  puts -nonewline stdout "\nOK to overwrite? (y) "
	  set answer [gets stdin]
	  if { [string match {[Nn]*} $answer] } {
	      error "$progname: Abort."
	  }
      }


      #
      # BEGIN:  Critical section
      #
      trap SIG_IGN SIGINT
      trap SIG_IGN SIGQUIT

      #
      # Deallocate private eeprom file for this domain, if it exists.
      #
      catch {free_eeprom_file $new_domain_name} status
      catch {rm_domain_history_entry $new_domain_name} status
  }

	#
	# Make sure domain is not up and running.  
	#

	set env(SUNW_HOSTNAME) $domain_name
	if [ catch { check_host -q } result ] {
		if { [string length $result] } {
			error "$progname: $result"
		} else {
			error "$progname: Error checking domain '$domain_name'"
		}
	}
	if { $result == 0 } {
		# Host is UP
		puts "\n$progname: Domain '$domain_name' is UP, must be shutdown before renaming."
 		error "$progname: exiting."
		}

  #
  # BEGIN:  Critical section
  #
  trap SIG_IGN SIGINT
  trap SIG_IGN SIGQUIT

  set e $env(SSPVAR)/.ssp_private/$EEPROM_DIR
  exec mv $e/eeprom.image.$domain_name $e/eeprom.image.$new_domain_name

  catch {dc_file_mgr "rename" [lindex $d_list $gdcfg_physical_position] $domain_name $new_domain_name} status

  #
  # Get SSP name from OS environment
  #
  set input [open "|uname -n" r]
  set ssp_name [gets $input]
  close $input

  #
  # Substitute new name into old list structure, then add to domain_config file.
  #
  catch {rm_domain_entry $domain_name} status
  catch {rm_ssp_to_domain_hosts_entry $domain_name} status

  # Fix for bugID: 4068732 - domain_rename destorys domain_config file when only one
  # domain entry is trying to be renamed.
  #
  # The problem is that the lreplace was not including the {}'s around the new
  # domain name entry. The fix makes sure that the {}'s are put in.
  #
  set new_dc_entry [list \
	[concat "{$new_domain_name}" [lrange $d_list 1 [llength $d_list]]]]
  if [ catch {put_domain_entry $new_dc_entry} status ] {
	 error "put_domain_entry : $status"
  }

  catch {put_ssp_to_domain_hosts_entry [list [list $new_domain_name $ssp_name]]} status

  # remove the old domain information from the ssp_resource file.
  catch { ssp_rm_resource $domain_name.domainInterruptVector }
  catch { ssp_rm_resource $domain_name.domainBootProc }
  catch { ssp_rm_resource $domain_name.domainSysBrdConfig }
  catch { ssp_rm_resource $domain_name.domainProcConfig }
  catch { ssp_rm_resource $domain_name.domainABusConfig }
  catch { ssp_rm_resource $domain_name.domainDBusConfig }

#
# kill obp_helper and netcon_server of this domain
#
  set obpfile $env(SSPVAR)/pids/obp_helper-$env(SUNW_HOSTNAME).pid
  set netfile $env(SSPVAR)/pids/netcon_server-$env(SUNW_HOSTNAME).pid

  if {[ tokenizeFile $obpfile tok " " ] == 1 } {
      set pid [lindex $tok(1) 0]
      catch { exec kill -s KILL $pid }
  }
  if {[ tokenizeFile $netfile tok " " ] == 1 } {
      set pid [lindex $tok(1) 0]
      catch { exec kill -s KILL $pid }
  }
	
  puts stdout "Domain : $domain_name is renamed to $new_domain_name !, \nNOTE: The domain boot disk name may also need to be changed"

  #
  # END:  Critical section
  #
  trap SIGINT
  trap SIGQUIT
}

##############################################################################
#
# domain_status_usage:  Print usage info for domain_status command and
#			raise error.
#
# inputs:
#	none
#
# outputs: 
#	none
#
##############################################################################
proc domain_status_usage {} {
  upvar progname progname
  error "Usage: $progname \[-d domain_name\]"
}

##############################################################################
#
# domain_status:  Print information on one or more active domains.
#
# inputs:
#	args	list of comand line arguments, passed as an array.
#
# outputs: 
#	0 on success, 1 on failure.
#
##############################################################################
proc domain_status { args } \
{
  global env
  global domain_tools_debug
  global gdcfg_name_position
  global gdcfg_sysbds_position
  global gdcfg_idninfo_position

  set progname "domain_status"
  set dcf $env(SSPVAR)/.ssp_private/domain_config

  check_ssp_env
  set_default_platform_type
  check_domain_history

  #
  # Default argument values.
  #
  set domain_name ""

  #
  # Find number of arguments, process/validate each.
  #
  set argcnt [llength $args]
  set ac 0
  while {$ac < $argcnt} {
    #
    # Process command arguments.  Each option must have an argument; if not,
    # print usage information and raise error.
    #
    switch -exact -- [lindex $args $ac] {
      -d { if {[incr ac] >= $argcnt} { dc_usage } ;
		set domain_name [lindex $args $ac] }
      default { domain_status_usage }
    }
    incr ac
  }

  if { ![file exists $env(SSPVAR)/.ssp_private/domain_config] } {
    error "$progname: No domains configured."
  }

  if { [string length $domain_name ] } {
    #
    # Check that domain entry exists.
    #
    set d_list [lindex [get_domain_entry $domain_name] 0]
    if { ![llength $d_list] } {
      error "$progname: No such domain: '$domain_name'."
    }
    #
    # Make list of 1 list to simplify algorithm loop below.
    #
    set d_config [list $d_list]
    set d_name [lindex $d_list $gdcfg_name_position];
  } else {
    #
    # Retrieve entire contents of domain_config file.
    #
    set d_config [get_domain_config]
    if { [llength $d_config] == 0 } {
        error "$progname: No domains configured."
    }
    set d_name {}
  }


  #
  # Loop through each domain specified and print information nicely.
  #
  set lastlength 14
  for {set i 0} {$i < [llength $d_config]} {incr i} {
    	set d_list [lindex $d_config $i]
    	set donamelength [ string length [lindex $d_list $gdcfg_name_position] ]
    	if { $donamelength > $lastlength } { set lastlength $donamelength }
#	Sort the board numbers before printing.
	set d_list [ lreplace $d_list $gdcfg_idninfo_position $gdcfg_idninfo_position \
			[ lsort -integer [lindex $d_list $gdcfg_sysbds_position] ] ]
	set f_list [ lappend f_list $d_list ]
  }
  # adjust length of domain name format on the fly
  set lastlength [ expr ($lastlength +2) ]
  set f "%-"
  append f $lastlength "s%-25s%-15s%-7s%s"
  
  #
  # Print header info and data
  #
  puts stdout [format $f DOMAIN TYPE PLATFORM OS SYSBDS]
    for {set i 0} {$i < [llength $f_list]} {incr i} {
      set d_list [lindex $f_list $i]
      puts stdout [eval format $f $d_list]
      }

  #
  # Dump IDN net status
  #
  idn_net_status $d_name
}

##############################################################################
#
# domain_history_usage:  Print usage info for domain_history command and
#			 raise error.
#
# inputs:
#	none
#
# outputs: 
#	none
#
##############################################################################
proc domain_history_usage {} {
  upvar progname progname
  error "Usage: $progname \[-d domain_name\] \[-r\]"
}

##############################################################################
#
# domain_history:  Access domain history file - print information on one or
#		   more domains, or remove one or all domain entries from
#		   the history file.
#
# inputs:
#	args	list of comand line arguments, passed as an array.
#
# outputs: 
#	0 on success, 1 on failure.
#
##############################################################################
proc domain_history { args } \
{
  global domain_tools_debug
  global env
  global gdcfg_name_position
  global gdcfg_sysbds_position
  global gdcfg_idninfo_position

  set progname "domain_history"

  check_user
  check_ssp_env
  set_default_platform_type
  check_domain_history

  #
  # Default argument values.
  #
  set domain_name ""
  set r_op 0

  #
  # Find number of arguments, process/validate each.
  #
  set argcnt [llength $args]
  set ac 0
  while {$ac < $argcnt} {
    #
    # Process command arguments.  -d must have an argument, -r does not.
    #
    switch -exact -- [lindex $args $ac] {
      -d { if {[incr ac] >= $argcnt} { domain_history_usage } ;
		set domain_name [lindex $args $ac] }
      -r { incr r_op }
      default { domain_history_usage }
    }
    incr ac
  }

  #
  # If domain was specified, list or remove that entry; if no domain was
  # specified, list or remove all entries.
  #
  if { [string length $domain_name ] } {
    if { $r_op } {
      free_eeprom_file $domain_name
      rm_domain_history_entry $domain_name
    } else {
      set d_list [lindex [get_domain_history_entry $domain_name] 0]
      if { [llength $d_list] } {
      puts stdout [format "%-15s%-25s%-15s%-7s%s" \
				DOMAIN TYPE PLATFORM OS SYSBDS]
#	Sort the board numbers before printing.
	set d_list [ lreplace $d_list $gdcfg_idninfo_position $gdcfg_idninfo_position \
			[ lsort -integer [lindex $d_list $gdcfg_sysbds_position] ] ]
        puts stdout [eval format "%-15s%-25s%-15s%-7s%s" $d_list]
      } else {	error "$progname: No such domain: '$domain_name'."
	}
    }
  } else {
    set h_llist [get_domain_history]
    if { $r_op } {
      free_eeprom_file ""
      #
      # Remove each entry in domain_history.
      #
      foreach entry $h_llist {
        rm_domain_history_entry [lindex $entry $gdcfg_name_position]
      }
    } else {
      #
      # Print header info
      #
      puts stdout [format "%-15s%-25s%-15s%-7s%s" \
				DOMAIN TYPE PLATFORM OS SYSBDS]
      for {set i 0} {$i < [llength $h_llist]} {incr i} {
        set d_list [lindex $h_llist $i]
#	Sort the board numbers before printing.
	set d_list [ lreplace $d_list $gdcfg_idninfo_position $gdcfg_idninfo_position \
			[ lsort -integer [lindex $d_list $gdcfg_sysbds_position] ] ]
        puts stdout [eval format "%-15s%-25s%-15s%-7s%s" $d_list]
      }
    }
  }
}

##############################################################################
#
# dc_file_mgr:  Domain config file management.  This procedure centralizes
#		manipulation of domain-specific files for domain_create
#		domain_remove and domain_rename commands.
#
# inputs:
#	op		"create", "remove", or "rename".  Determines the file
#			operations to be performed.
#	p_name		platform name
#	d_name		domain name
#	new_d_name	(optional) new domain name
#	p_type		(optional) platform type
#
# outputs: 
#	none
#
##############################################################################
proc dc_file_mgr { op p_name d_name {new_d_name none} {p_type DEFAULT_TYPE} } \
{
  global env
  global D_PLATFORM_TYPE
  global domain_tools_debug
  upvar progname progname

  set PLATFORM_ERC_FILE edd.platform.erc
  set PLATFORM_EMC_FILE edd.emc
  set DOMAIN_ERC_FILE edd.domain.erc
  set EDD_ERC_FILE edd.erc

  if {$domain_tools_debug > 0} {
    puts stdout "dc_file_mgr: '$op' '$p_name' '$d_name' '$p_type' '$new_d_name'"
  }

  #
  # The following was necessary because we can't embed something like
  # $D_PLATFORM_TYPE in a default argument definition.
  #
  if { [string match $p_type DEFAULT_TYPE] } {
    set p_type $D_PLATFORM_TYPE
  }

  set adm $env(SSPVAR)/adm
  set d_adm $adm/$d_name
  set p_etc $env(SSPVAR)/etc/$p_name
  set d_etc $p_etc/$d_name
  set tmpl $env(SSPVAR)/.ssp_private/templates/$p_type
  set d_config $env(SSPVAR)/.ssp_private/domain_config

  if { [string compare $op "create"] == 0 } {
    #
    # Create required files and directories.
    #
    if { ![file exists $d_adm/post] } {
      exec mkdir -p $d_adm/post
    } else {
      puts "Note: directory $d_adm already exists"
    }
    if { ![file exists $d_etc] } {
      exec mkdir -p $d_etc
    } else {
      puts "Note: directory $d_etc already exists"
    }
    if { ![file exists $p_etc/$EDD_ERC_FILE] } {
      exec cp $tmpl/$PLATFORM_ERC_FILE $p_etc/$EDD_ERC_FILE
    }
    if { ![file exists $p_etc/$PLATFORM_EMC_FILE] } {
      exec cp $tmpl/$PLATFORM_EMC_FILE $p_etc/$PLATFORM_EMC_FILE
    }
    if { [file exists $tmpl/$DOMAIN_ERC_FILE] } {
      if { ![file exists $d_etc/$EDD_ERC_FILE] } {
         exec cp $tmpl/$DOMAIN_ERC_FILE $d_etc/$EDD_ERC_FILE
      }
    } else {
      puts stderr "$progname: Warning! Could not find $tmpl/$DOMAIN_ERC_FILE."
    }
  } elseif { [string compare $op "remove"] == 0 } {
    #
    # Remove domain-associated files and directories.
    #
    exec rm -rf $d_adm
    exec rm -rf $d_etc

  } elseif { [string compare $op "rename"] == 0 } {
    #
    # Rename domain-associated files and directories.
    #
    exec mv $p_etc/$d_name $p_etc/$new_d_name
    exec mv $adm/$d_name $adm/$new_d_name
  } else {
    error "$progname: Unknown file operation: '$op'"
  }
}

#
# idn_net_status: dumps the IDNnet configuration from the MIB
#
proc idn_net_status { d_name } \
{
	upvar handle handle
	upvar progname progname

	if { [ llength $d_name ] > 0 } {
		set idn_config [ get_idn_config $d_name ]
	} else {
		set idn_config [ get_idn_config ]
	}

	if { [ llength $idn_config ] > 0 } {
		puts stdout "\nIDN NETWORKS"
	}
	for { set i 0 } { $i < [ llength $idn_config ] } { incr i } {
		set idn_entry [ lindex $idn_config $i ]
		set idn_id [ lindex $idn_entry 0 ]
		set idn_members [ lsort -ascii [ lindex $idn_entry 1 ] ]

		set mlist {}
		set mcount [ expr [ llength $idn_members ] - 1 ]
		for { set j 0 } { $j < $mcount } { incr j } {
			lappend mlist [ format "%s," [ lindex $idn_members $j ] ]
		}
		lappend mlist [ lindex $idn_members $j ]
		puts stdout "$idn_id: $mlist"
	}
}
